'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var prettier = require('prettier');

function isASTNode(n) {
    return 'html' in n && 'tokens' in n;
}

function extractAttributes(html) {
    const extractAttributesRegex = /<[a-z]+\s*(.*?)>/i;
    const attributeRegex = /([^\s=]+)(?:=("|')(.*?)\2)?/gi;
    const [, attributesString] = html.match(extractAttributesRegex);
    const attrs = [];
    let match;
    while ((match = attributeRegex.exec(attributesString))) {
        const [all, name, quotes, value] = match;
        const attrStart = match.index;
        let valueNode;
        if (!value) {
            valueNode = true;
        }
        else {
            let valueStart = attrStart + name.length;
            if (quotes) {
                valueStart += 2;
            }
            valueNode = [
                {
                    type: 'Text',
                    data: value,
                    start: valueStart,
                    end: valueStart + value.length,
                },
            ];
        }
        attrs.push({
            type: 'Attribute',
            name,
            value: valueNode,
            start: attrStart,
            end: attrStart + all.length,
        });
    }
    return attrs;
}

function getText(node, options) {
    return options.originalText.slice(options.locStart(node), options.locEnd(node));
}

function isBindingNodeV2(node) {
    return node.type === 'Binding' && 'value' in node;
}

function getSvelteVersion(importPath) {
    const { version } = require(`${importPath}/package.json`);
    const [major, minor, misc] = version.split('.');
    const [patch, label] = misc.split('-');
    return {
        major: Number(major),
        minor: Number(minor),
        patch: Number(patch),
        label,
    };
}

const { concat, join, line, group, indent, softline, hardline } = prettier.doc.builders;
function print(path, options, print) {
    const n = path.getValue();
    if (!n) {
        return '';
    }
    if (isASTNode(n)) {
        const parts = [path.call(print, 'html')];
        if (n.css) {
            n.css.type = 'Style';
            n.css.content.type = 'StyleProgram';
            parts.push(path.call(print, 'css'));
        }
        if (n.js) {
            n.js.type = 'Script';
            parts.push(path.call(print, 'js'));
        }
        if (n.instance) {
            n.instance.type = 'Script';
            n.instance.attributes = extractAttributes(getText(n.instance, options));
            parts.push(path.call(print, 'instance'));
        }
        if (n.module) {
            n.module.type = 'Script';
            n.module.attributes = extractAttributes(getText(n.module, options));
            parts.push(path.call(print, 'module'));
        }
        return group(join(hardline, parts));
    }
    const version = getSvelteVersion(options.sveltePath);
    let [open, close] = ['{', '}'];
    if (version.major < 3) {
        [open, close] = ['"', '"'];
    }
    const node = n;
    switch (node.type) {
        case 'Fragment':
            const children = node.children;
            return concat([
                printChildren(path, print, {
                    skipFirst: true,
                    filter: (node, i) => {
                        if (i === 0 && node.type === 'Text' && node.data.trim() === '') {
                            return false;
                        }
                        let include = false;
                        for (let j = i; j < children.length; j++) {
                            const child = children[j];
                            if (!(child.type === 'Text' && child.data.trim() === '')) {
                                include = true;
                                break;
                            }
                        }
                        return include;
                    },
                }),
                hardline,
            ]);
        case 'Text':
            return join(hardline, node.data
                .replace(/\n(?!\n)/g, '')
                .replace(/[ \t]+/g, ' ')
                .split(/\n/g));
        case 'Element':
        case 'InlineComponent':
        case 'Slot':
        case 'Window':
        case 'Head':
        case 'Title':
            return group(concat([
                '<',
                node.name,
                indent(group(concat([
                    node.type === 'InlineComponent' && node.expression
                        ? concat([
                            line,
                            'this={',
                            printJS(path, print, 'expression'),
                            '}',
                        ])
                        : '',
                    ...path.map(childPath => childPath.call(print), 'attributes'),
                ]))),
                node.children.length ? '>' : ' />',
                indent(printChildren(path, print)),
                node.children.length ? concat([softline, '</', node.name, '>']) : '',
            ]));
        case 'Identifier':
            return node.name;
        case 'Attribute': {
            const hasLoneMustacheTag = node.value !== true &&
                node.value.length === 1 &&
                node.value[0].type === 'MustacheTag';
            let isAttributeShorthand = node.value !== true &&
                node.value.length === 1 &&
                node.value[0].type === 'AttributeShorthand';
            // Convert a={a} into {a}
            if (hasLoneMustacheTag) {
                const expression = node.value[0].expression;
                isAttributeShorthand =
                    expression.type === 'Identifier' && expression.name === node.name;
            }
            if (isAttributeShorthand) {
                return concat([line, '{', node.name, '}']);
            }
            const def = [line, node.name];
            if (node.value !== true) {
                def.push('=');
                const quotes = !hasLoneMustacheTag;
                quotes && def.push('"');
                def.push(...path.map(childPath => childPath.call(print), 'value'));
                quotes && def.push('"');
            }
            return concat(def);
        }
        case 'MustacheTag':
            return concat(['{', printJS(path, print, 'expression'), '}']);
        case 'IfBlock': {
            const def = [
                '{#if ',
                printJS(path, print, 'expression'),
                '}',
                indent(printChildren(path, print)),
                line,
            ];
            if (node.else) {
                def.push(path.call(print, 'else'));
            }
            def.push('{/if}');
            return group(concat(def));
        }
        case 'ElseBlock': {
            // Else if
            if (node.children.length === 1 && node.children[0].type === 'IfBlock') {
                const ifNode = node.children[0];
                const def = [
                    '{:elseif ',
                    path.map(ifPath => printJS(path, print, 'expression'), 'children')[0],
                    '}',
                    indent(path.map(ifPath => printChildren(ifPath, print), 'children')[0]),
                    line,
                ];
                if (ifNode.else) {
                    def.push(path.map(ifPath => ifPath.call(print, 'else'), 'children')[0]);
                }
                return group(concat(def));
            }
            return group(concat(['{:else}', indent(printChildren(path, print)), line]));
        }
        case 'EachBlock': {
            const def = [
                '{#each ',
                printJS(path, print, 'expression'),
                ' as ',
                printJS(path, print, 'context'),
            ];
            if (node.index) {
                def.push(', ', node.index);
            }
            if (node.key) {
                def.push(' (', printJS(path, print, 'key'), ')');
            }
            def.push('}', indent(printChildren(path, print)), line);
            if (node.else) {
                def.push(path.call(print, 'else'));
            }
            def.push('{/each}');
            return group(concat(def));
        }
        case 'AwaitBlock': {
            return group(concat([
                group(concat(['{#await ', printJS(path, print, 'expression'), '}'])),
                indent(path.call(print, 'pending')),
                line,
                group(concat(['{:then', node.value ? ' ' + node.value : '', '}'])),
                indent(path.call(print, 'then')),
                line,
                group(concat(['{:catch', node.error ? ' ' + node.error : '', '}'])),
                indent(path.call(print, 'catch')),
                line,
                '{/await}',
            ]));
        }
        case 'ThenBlock':
        case 'PendingBlock':
        case 'CatchBlock':
            return printChildren(path, print);
        case 'EventHandler':
            return concat([
                line,
                'on:',
                node.name,
                node.expression
                    ? concat(['=', open, printJS(path, print, 'expression'), close])
                    : '',
            ]);
        case 'Binding':
            if (isBindingNodeV2(node)) {
                return concat([
                    line,
                    'bind:',
                    node.name,
                    node.value.type === 'Identifier' && node.value.name === node.name
                        ? ''
                        : concat(['=', open, printJS(path, print, 'value'), close]),
                ]);
            }
            return concat([
                line,
                'bind:',
                node.name,
                node.expression.type === 'Identifier' && node.expression.name === node.name
                    ? ''
                    : concat(['=', '{', printJS(path, print, 'expression'), '}']),
            ]);
        case 'DebugTag':
            return concat([
                '{@debug',
                node.identifiers.length > 0
                    ? concat([' ', join(', ', path.map(print, 'identifiers'))])
                    : '',
                '}',
            ]);
        case 'Ref':
            return concat([line, 'ref:', node.name]);
        case 'Comment':
            return group(concat(['<!--', node.data, '-->']));
        case 'Transition':
            const kind = node.intro && node.outro ? 'transition' : node.intro ? 'in' : 'out';
            return concat([
                line,
                kind,
                ':',
                node.name,
                node.expression
                    ? concat(['=', open, printJS(path, print, 'expression'), close])
                    : '',
            ]);
        case 'Action':
            return concat([
                line,
                'use:',
                node.name,
                node.expression
                    ? concat(['=', open, printJS(path, print, 'expression'), close])
                    : '',
            ]);
        case 'Animation':
            return concat([
                line,
                'animate:',
                node.name,
                node.expression
                    ? concat(['=', open, printJS(path, print, 'expression'), close])
                    : '',
            ]);
        case 'RawMustacheTag':
            return concat(['{@html ', printJS(path, print, 'expression'), '}']);
        case 'Spread':
            return concat([line, '{...', printJS(path, print, 'expression'), '}']);
    }
    console.log(JSON.stringify(node, null, 4));
    throw new Error('unknown node type: ' + node.type);
}
function printChildren(path, print, { skipFirst = false, filter = (node, i) => true } = {}) {
    const children = [];
    let i = 0;
    let isFirst = true;
    const childNodes = path.getValue().children;
    path.each(childPath => {
        const child = childPath.getValue();
        const index = i;
        i++;
        if (!filter(child, index)) {
            return;
        }
        if (!(isFirst && skipFirst)) {
            if (isInlineNode(child)) {
                if (!isEmptyNode(child)) {
                    let lineType = softline;
                    if (child.type === 'Text') {
                        if (/^\s+/.test(child.data)) {
                            // Remove leading spaces
                            child.data = trimStart(child.data);
                            if (!isFirst) {
                                lineType = line;
                            }
                        }
                    }
                    children.push(lineType);
                }
            }
            else {
                children.push(hardline);
            }
        }
        if (child.type === 'Text') {
            if (isLastNode(childNodes, filter, index)) {
                // Remove trailing spaces
                child.data = trimEnd(child.data);
            }
        }
        children.push(childPath.call(print));
        isFirst = false;
    }, 'children');
    return concat(children);
}
function printJS(path, print, name) {
    if (!name) {
        path.getValue().isJS = true;
        return path.call(print);
    }
    path.getValue()[name].isJS = true;
    return path.call(print, name);
}
function isInlineNode(node) {
    return node.type === 'Text' || node.type === 'MustacheTag';
}
function isEmptyNode(node) {
    return node.type === 'Text' && node.data.trim() === '';
}
function isLastNode(nodes, filter, index) {
    for (let i = index + 1; i < nodes.length; i++) {
        if (filter(nodes[i], i)) {
            return false;
        }
    }
    return true;
}
function trimStart(text) {
    return text.replace(/^\s+/, '');
}
function trimEnd(text) {
    return text.replace(/\s+$/, '');
}

const { builders: { concat: concat$1, hardline: hardline$1, group: group$1, indent: indent$1 }, utils: { removeLines }, } = prettier.doc;
function embed(path, print, textToDoc, options) {
    const node = path.getNode();
    if (node.isJS) {
        return removeLines(textToDoc(getText(node, options), {
            parser: expressionParser,
            singleQuote: true,
        }));
    }
    switch (node.type) {
        case 'Script':
            return embedTag(path, print, textToDoc, node, 'babel');
        case 'Style':
            return embedTag(path, print, textToDoc, node, 'css');
    }
    return null;
}
function expressionParser(text, parsers) {
    const ast = parsers.babylon(`(${text})`);
    return {
        type: 'File',
        program: ast.program.body[0].expression,
    };
}
function skipBlank(docs) {
    for (let i = docs.length - 1; i >= 0; i--) {
        const doc = docs[i];
        if (typeof doc !== 'string') {
            if (doc.type === 'break-parent') {
                continue;
            }
        }
        return i;
    }
    return -1;
}
function nukeLastLine(doc) {
    if (typeof doc === 'string') {
        return doc;
    }
    switch (doc.type) {
        case 'concat':
            const end = skipBlank(doc.parts);
            if (end > -1) {
                return concat$1([
                    ...doc.parts.slice(0, end),
                    nukeLastLine(doc.parts[end]),
                    ...doc.parts.slice(end + 1),
                ]);
            }
            break;
        case 'line':
            return '';
    }
    return doc;
}
function embedTag(path, print, textToDoc, node, parser) {
    const contentAttribute = node.attributes.find(n => n.name === '✂prettier:content✂');
    let content = '';
    if (contentAttribute &&
        Array.isArray(contentAttribute.value) &&
        contentAttribute.value.length > 0) {
        const encodedContent = contentAttribute.value[0].data;
        content = Buffer.from(encodedContent, 'base64').toString('ascii');
    }
    node.attributes = node.attributes.filter(n => n !== contentAttribute);
    return group$1(concat$1([
        '<',
        node.type.toLowerCase(),
        indent$1(group$1(concat$1(path.map(childPath => childPath.call(print), 'attributes')))),
        '>',
        indent$1(concat$1([hardline$1, nukeLastLine(textToDoc(content, { parser }))])),
        hardline$1,
        '</',
        node.type.toLowerCase(),
        '>',
        hardline$1,
    ]));
}

function snipTagContent(tagName, source, placeholder = '') {
    const regex = new RegExp(`<${tagName}([^]*?)>([^]*?)<\/${tagName}>`, 'gi');
    return source.replace(regex, (_, attributes, content) => {
        const encodedContext = Buffer.from(content).toString('base64');
        return `<${tagName}${attributes} ✂prettier:content✂="${encodedContext}">${placeholder}</${tagName}>`;
    });
}

function importSvelte(importPath) {
    const version = getSvelteVersion(importPath);
    if (version.major <= 2) {
        return require(importPath);
    }
    const svelte = require(`${importPath}/compiler`);
    return {
        parse(text) {
            return svelte.compile(text, { generate: false, onwarn: () => undefined }).ast;
        },
    };
}

function locStart(node) {
    return node.start;
}
function locEnd(node) {
    return node.end;
}
const languages = [
    {
        name: 'svelte',
        parsers: ['svelte'],
        extensions: ['.svelte', '.html'],
    },
];
const parsers = {
    svelte: {
        parse: (text, parsers, options) => {
            const { parse } = importSvelte(options.sveltePath);
            return parse(text);
        },
        preprocess(text) {
            text = snipTagContent('style', text);
            text = snipTagContent('script', text, '{}');
            return text;
        },
        locStart,
        locEnd,
        astFormat: 'svelte-ast',
    },
};
const printers = {
    'svelte-ast': {
        print,
        embed,
    },
};
const options = {
    sveltePath: {
        type: 'path',
        description: 'Path to custom Svelte version to use',
        default: 'svelte',
    },
};

exports.languages = languages;
exports.parsers = parsers;
exports.printers = printers;
exports.options = options;
//# sourceMappingURL=plugin.js.map
